package cz.cuni.amis.pogamut.ut2004.lectures.hsbot;

import java.util.Collection;
import java.util.logging.Level;

import cz.cuni.amis.pathfinding.alg.astar.AStarResult;
import cz.cuni.amis.pathfinding.map.IPFMapView;
import cz.cuni.amis.pogamut.base.agent.navigation.impl.PrecomputedPathFuture;
import cz.cuni.amis.pogamut.base.utils.guice.AgentScoped;
import cz.cuni.amis.pogamut.base3d.worldview.object.ILocated;
import cz.cuni.amis.pogamut.ut2004.bot.impl.UT2004Bot;
import cz.cuni.amis.pogamut.ut2004.communication.messages.gbcommands.Initialize;
import cz.cuni.amis.pogamut.ut2004.communication.messages.gbinfomessages.BotKilled;
import cz.cuni.amis.pogamut.ut2004.communication.messages.gbinfomessages.ConfigChange;
import cz.cuni.amis.pogamut.ut2004.communication.messages.gbinfomessages.GameInfo;
import cz.cuni.amis.pogamut.ut2004.communication.messages.gbinfomessages.InitedMessage;
import cz.cuni.amis.pogamut.ut2004.communication.messages.gbinfomessages.NavPoint;
import cz.cuni.amis.pogamut.ut2004.hideandseek.bot.UT2004BotHSController;
import cz.cuni.amis.pogamut.ut2004.hideandseek.protocol.HSBotState;
import cz.cuni.amis.pogamut.ut2004.hideandseek.protocol.HSScoreChangeReason;
import cz.cuni.amis.pogamut.ut2004.hideandseek.protocol.messages.HSAssignSeeker;
import cz.cuni.amis.pogamut.ut2004.hideandseek.protocol.messages.HSBotStateChanged;
import cz.cuni.amis.pogamut.ut2004.hideandseek.protocol.messages.HSPlayerScoreChanged;
import cz.cuni.amis.pogamut.ut2004.hideandseek.protocol.messages.HSRoundEnd;
import cz.cuni.amis.pogamut.ut2004.hideandseek.protocol.messages.HSRoundStart;
import cz.cuni.amis.pogamut.ut2004.hideandseek.protocol.messages.HSRoundState;
import cz.cuni.amis.pogamut.ut2004.hideandseek.protocol.messages.HSRunnerCaptured;
import cz.cuni.amis.pogamut.ut2004.hideandseek.protocol.messages.HSRunnerFouled;
import cz.cuni.amis.pogamut.ut2004.hideandseek.protocol.messages.HSRunnerSafe;
import cz.cuni.amis.pogamut.ut2004.hideandseek.protocol.messages.HSRunnerSpotted;
import cz.cuni.amis.pogamut.ut2004.hideandseek.protocol.messages.HSRunnerSurvived;
import cz.cuni.amis.pogamut.ut2004.hideandseek.protocol.messages.HSSeekerFouled;
import cz.cuni.amis.pogamut.ut2004.utils.UT2004BotRunner;
import cz.cuni.amis.utils.IFilter;
import cz.cuni.amis.utils.collections.MyCollections;
import cz.cuni.amis.utils.exception.PogamutException;

/**
 * HideAndSeek bot torso
 *
 * @author Jimmy
 */
@AgentScoped
public class HideAndSeekBot extends UT2004BotHSController<UT2004Bot> {

    /**
     * Just for the numbering of bots.
     */
    private static int number = 0;
   
    /**
     * Here we can modify initializing command for our bot.
     *
     * @return
     */
    @Override
    public Initialize getInitializeCommand() {
        // TODO: substitute "PutYourNameHere" with your full name, e.g. "JakubGemrot"
        return new Initialize().setName("PutYourNameHere" + (++number));
    }

    /**
     * The bot is initialized in the environment - a physical representation of
     * the bot is present in the game.
     *
     * @param config information about configuration
     * @param init information about configuration
     */
    @SuppressWarnings("unchecked")
    @Override
    public void botInitialized(GameInfo gameInfo, ConfigChange config, InitedMessage init) {
    	bot.getLogger().getCategory("Yylex").setLevel(Level.OFF);
        navigation.setLogLevel(Level.WARNING);
        nmNav.setLogLevel(Level.WARNING);
    }

    // --------------------
    // ====================
    // HIDE AND SEEK EVENTS
    // ====================
    // --------------------
    /**
     * Some RUNNER has just survived the round.
     * <p>
     * <p>
     * This event is triggered at the end of the round for every RUNNER that has
     * not been FOULED, CAPTURED and did not make it to SAFE area.
     *
     * @param event
     */
    @Override
    protected void hsRunnerSurvived(HSRunnerSurvived event, boolean me) {
    }

    /**
     * Some RUNNER has just been spotted by the SEEKER.
     *
     * @param event
     */
    @Override
    protected void hsRunnerSpotted(HSRunnerSpotted event, boolean me) {
    }

    /**
     * Some RUNNER has just made it to the safe-area.
     *
     * @param event
     */
    @Override
    protected void hsRunnerSafe(HSRunnerSafe event, boolean me) {
    }

    /**
     * Some RUNNER has just been fouled-out because it stepped into the
     * restricted-area while it was activated.
     *
     * @param event
     * @param me whether this event speaks about your bot
     */
    @Override
    protected void hsRunnerFouled(HSRunnerFouled event, boolean me) {
    }

    /**
     * Some RUNNER has just been captured by the SEEKER, i.e., seeker has
     * spotted the runner and made it to the safe-area before the runner.
     *
     * @param event
     * @param me whether this event speaks about your bot
     */
    @Override
    protected void hsRunnerCaptured(HSRunnerCaptured event, boolean me) {
    }

    /**
     * Round state update has been received.
     *
     * @param event
     */
    @Override
    protected void hsRoundState(HSRoundState event) {
    }

    /**
     * New round has just started, you may use this event to initialize your
     * data-structures.
     *
     * @param event
     */
    @Override
    protected void hsRoundStart(HSRoundStart event) {
        navigation.stopNavigation(); // prevent commands from previous round to be still effective

        // TODO: reset your state here		
    }

    /**
     * Round has just ended, you may use this event to cleanup your
     * data-structures.
     *
     * @param event
     */
    @Override
    protected void hsRoundEnd(HSRoundEnd event) {
    }

    /**
     * Some player (runner/seeker) score has just changed, you may examine the
     * reason within {@link HSPlayerScoreChanged#getScoreChangeReason()}, see
     * {@link HSScoreChangeReason}.
     *
     * @param event
     * @param me whether this event speaks about your bot
     */
    @Override
    protected void hsPlayerScoreChanged(HSPlayerScoreChanged event, boolean me) {
    }

    /**
     * Some bot {@link HSBotState} has been updated.
     *
     * @param event
     * @param me whether this event speaks about your bot
     */
    protected void hsBotStateChanged(HSBotStateChanged event, boolean me) {
    }

    /**
     * SEEKER role has been just assigned to some bot.
     *
     * @param event
     * @param me whether this event speaks about your bot
     */
    @Override
    protected void hsAssignSeeker(HSAssignSeeker event, boolean me) {
    }
    
    /**
     * SEEKER has been fould out due to dwelling within restricted area for too long.
     */
    @Override
    protected void hsSeekerFouled(HSSeekerFouled event, boolean me) {    
    }

    /**
     * This method is called only once right before actual logic() method is
     * called for the first time.
     */
    @Override
    public void beforeFirstLogic() {
    }

    /**
     * Main method that controls the bot - makes decisions what to do next. It
     * is called iteratively by Pogamut engine every time a synchronous batch
     * from the environment is received. This is usually 4 times per second - it
     * is affected by visionTime variable, that can be adjusted in GameBots ini
     * file in UT2004/System folder.
     */
    @Override
    public void logic() {
        // mark that another logic iteration has began
        log.info("====== LOGIC ITERATION ======");

        // UNCOMMENT TO GET DETAIL INFO DURING RUN-TIME
        // logState();
        if (!hide.isRoundRunning()) {
            // Game or Game-Round is not running yet...
            return;
        }
        
        if (!hide.isMeAlive()) {
            // I'm not alive ~ already captured / escaped
            return;
        }

        switch (hide.getMyState()) {
            case RUNNER:
                logicRunner();
                break;
            case RUNNER_SPOTTED:
                logicRunnerSpotted();
                break;
            case SEEKER:
                logicSeeker();
                break;
            default:
                // SHOULD NOT REACH HERE, BUT LET'S PLAY SAFE...
                return;
        }
    }

    private void logState() {
        log.info("------      STATE      ------");
        log.info("Game state:                  " + hide.getGameState());
        log.info("Game running:                " + hide.isGameRunning());
        if (!hide.isGameRunning()) {
            return;
        }
        log.info("Round running:               " + hide.isRoundRunning());
        if (!hide.isRoundRunning()) {
            return;
        }
        log.info("Me:                          " + hide.getMyState());
        log.info("Me spawned?                  " + hide.isMeAlive());
        log.info("Hiding time left:            " + hide.getRemainingHidingTime());
        log.info("Restricted area time left:   " + hide.getRemainingRestrictedAreaTime());
        log.info("Round time left:             " + hide.getRemainingRoundTime());
        log.info("Distance to safe-area:       " + hide.getMySafeAreaDistance());
        log.info("Distance to restricted-area: " + hide.getMyRestrictedAreaDistance());
    }

    // ------------
    // ============
    // LOGIC RUNNER
    // ============
    // ------------
    private void logicRunner() {
        log.info("------ RUNNER ------");
    	// YOU ARE RUNNER => one who should be hiding
        
        if (hide.isHidingTime()) {
            log.info("HIDING TIME");
            // SEEKER has not been spawned yet, you have time to hide out!    		
            if (!navigation.isNavigating()) {
                navigation.navigate(
                    MyCollections.getRandom(
                        MyCollections.getFiltered(
                            navPoints.getNavPoints().values(),
                            new IFilter<NavPoint>() {
                                @Override
                                public boolean isAccepted(NavPoint nav) {
                                    return hide.getSafeAreaDistance(nav) > 500;
                                }
                            })
                       )
                );
            }
            return;
        }

        if (hide.isRestrictedAreaActivated()) {
            log.info("RESTRICTED AREA ACTIVE");
    		// SEEKER has been spawned into the environment but you cannot save yourself yet
            //         "easy-win" by standing in the safe are at the beginning of the game is not allowed 
            return;
        }

        log.info("SAFE AREA ACTIVE");
    	// SEEKER is present in the game, but restricted area is not active, 
        // you can save yourself by running to safe area before the SEEKER spots you and returns first
        navigation.navigate(hide.getSafeArea());
        
        //NavPoint safeArea = info.getNearestNavPoint(hide.getSafeArea());
    }

    // --------------------
    // ====================
    // LOGIC RUNNER SPOTTED
    // ====================
    // --------------------
    private void logicRunnerSpotted() {
        log.info("------ RUNNER SPOTTED ------");
    	// YOU ARE RUNNER who has already been spotted by the SEEKER
        // if SEEKER gets to the safe point first, it will catch you
    }

    // ------------
    // ============
    // LOGIC SEEKER
    // ============
    // ------------
    private void logicSeeker() {
        log.info("------ SEEKER ------");
    	// you are SEEKER, one who should be finding other runners and catching them

        if (hide.canCaptureRunner()) {
            log.info("CAN CAPTURE RUNNER!");
            // you have had spotted some runner that have not made it to the safe area yet
            nmNav.navigate(hide.getSafeArea());
        } else {
    		// you cannot capture any runner by returning to the safe area,
            // you should try to spot some first    
            if (!nmNav.isNavigating()) nmNav.navigate(navPoints.getRandomNavPoint());
        }
        
    }
    
    // =====
    // AStar
    // =====
    
    private NavPoint lastAStarTarget = null;
    
    public boolean navigateAStarPath(NavPoint targetNavPoint) {
        if (lastAStarTarget == targetNavPoint) {
            if (navigation.isNavigating()) return true;
        }
        PrecomputedPathFuture<ILocated> path = getAStarPath(targetNavPoint);
        if (path == null) {
            navigation.stopNavigation();
            return false;
        }
        lastAStarTarget = targetNavPoint;
        navigation.navigate(path);
        return true;
    }
    
    private IPFMapView<NavPoint> mapView = new IPFMapView<NavPoint>() {

        @Override
        public Collection<NavPoint> getExtraNeighbors(NavPoint node, Collection<NavPoint> mapNeighbors) {
            return null;
        }

        @Override
        public int getNodeExtraCost(NavPoint node, int mapCost) {
            return 0;
        }

        @Override
        public int getArcExtraCost(NavPoint nodeFrom, NavPoint nodeTo, int mapCost) {
            return 0;
        }

        @Override
        public boolean isNodeOpened(NavPoint node) {
            return true;
        }

        @Override
        public boolean isArcOpened(NavPoint nodeFrom, NavPoint nodeTo) {
            return true;
        }
    };
    
    private PrecomputedPathFuture<ILocated> getAStarPath(NavPoint targetNavPoint) {
        NavPoint startNavPoint = info.getNearestNavPoint();
        AStarResult<NavPoint> result = aStar.findPath(startNavPoint, targetNavPoint, mapView);
        if (result == null || !result.isSuccess()) return null;
        PrecomputedPathFuture path = new PrecomputedPathFuture(startNavPoint, targetNavPoint, result.getPath());
        return path;
    }
    
    // ========
    // CLEAN UP
    // ========

    @Override
    public void botKilled(BotKilled event) {
        navigation.stopNavigation();
        lastAStarTarget = null;
    }

    public static void main(String args[]) throws PogamutException {
    	// LEAVE THIS BE!
        // Start the debug game via HideAndSeekGame !!!
        new UT2004BotRunner(HideAndSeekBot.class, "HSBot").setMain(true).setLogLevel(Level.WARNING).startAgents(1);
    }
}
